package br.ufrj.aula13

import scanner._    

object parser {
  // Trait para poder especificar parsers usando for
  trait Parser[T] {
    def parse(input: Stream[(Token, Int)]): (T, Stream[(Token, Int)]);
    
    def flatMap[U](f: T => Parser[U]): Parser[U] = {
      val p = this
      new Parser[U] {
        override def parse(input: Stream[(Token, Int)]): (U, Stream[(Token, Int)]) = {
          val (res, tail) = p.parse(input)
          f(res).parse(tail)
        }
      }
    }
  
    def map[U](f: T => U): Parser[U] = {
      val p = this
      new Parser[U] {
        override def parse(input: Stream[(Token, Int)]): (U, Stream[(Token, Int)]) = {
          val (res, tail) = p.parse(input)
          (f(res), tail)
        }
      }
    }
    
    def withFilter(f: T => Boolean): Parser[T] = {
      val p = this
      new Parser[T] {
        override def parse(input: Stream[(Token, Int)]): (T, Stream[(Token, Int)]) = {
          val (res, tail) = p.parse(input)
          if (f(res)) (res, tail) else sys.error("predicado falhou")
        }
      }
    }
  }

  // Precedência de operadores binários
  val prec: Map[Token, (Int, Int)] = Map(
    OP('|') -> (3, 3),
    OP('&') -> (4, 4),
    OP('<') -> (5, 5),
    EQ()    -> (5, 5),
    OP('+') -> (6, 6),
    OP('-') -> (6, 6),
    OP('*') -> (7, 7),
    OP('/') -> (7, 7),
    OP('=') -> (8, 2),
    OP(';') -> (2, 1)
  )
  
  // Construtores dos operadores binários
  val binExp: Map[Token, (interp.Exp, interp.Exp) => interp.Exp] = Map(
    OP('|') -> interp.Ou,
    OP('&') -> interp.E,
    OP('<') -> interp.Menor,
    OP('+') -> interp.Soma,
    OP('-') -> interp.Sub,
    OP('*') -> interp.Mult,
    OP('/') -> interp.Div,
    OP('=') -> interp.Atrib,
    OP(';') -> interp.Seq,
    EQ()    -> interp.Eq
  )
  
  // Próximo token tem que ser tok ou é erro
  def token(tok: Token) = new Parser[Token] {
    override def parse(input: Stream[(Token, Int)]): (Token, Stream[(Token, Int)]) =
      input.head match {
        case (token, line) if token == tok => (token, input.tail)
        case (token, line) => sys.error("erro de sintaxe na linha " + line +
            ", encontrado: " + token + ", esperado: " + tok)
      }
  }
  
  // Retorna próximo token sem avançar
  object peek extends Parser[Token] {
    override def parse(input: Stream[(Token, Int)]): (Token, Stream[(Token, Int)]) = {
      val (token, _) = input.head
      (token, input)
    }
  }

  // Erro de parsing
  def parseError[T](esperado: String) = new Parser[T] {
    override def parse(input: Stream[(Token, Int)]): (T, Stream[(Token, Int)]) = {
      val (token, line) = input.head
      sys.error("erro de sintaxe na linha: " + line + ", encontrado: " +
          token + ", esperado: " + esperado)
    }
  }

  // Introduz um valor no parser
  def unit[T](x: T) = new Parser[T] {
    override def parse(input: Stream[(Token, Int)]): (T, Stream[(Token, Int)]) =
      (x, input)
  }

  // Salta próximo token e introduz um valor
  def skip[T](x: T) = new Parser[T] {
    override def parse(input: Stream[(Token, Int)]): (T, Stream[(Token, Int)]) =
      (x, input.tail)
  }
  
  // Uma lista de ps separados por vírgula
  def listOf[T](p: Parser[T]): Parser[List[T]] = for {
    first <- p
    tok <- peek
    rest <- if (tok == OP(',')) (for {
                                   _ <- token(OP(','))
                                   rest <- listOf(p) 
                                 } yield rest)
            else unit(List[T]())
  } yield first :: rest

  // Identificadores
  object id extends Parser[String] {
    override def parse(input: Stream[(Token, Int)]): (String, Stream[(Token, Int)]) =
      input.head match {
        case (ID(s), line) => (s, input.tail)
        case (token, line) => sys.error("erro de sintaxe na linha " + line +
            ", encontrado: " + token + ", esperado identificador")
      }
  }  
  
  // Funções top-level
  val fun: Parser[interp.Fun1] = for {
    _ <- token(FUN())
    nome <- id
    _ <- token(OP('('))
    tok <- peek
    params <- if (tok == OP(')')) unit(List[String]()) 
              else listOf(id)
    _ <- token(OP(')'))
    corpo <- exp
    _ <- token(END())
  } yield interp.Fun1(nome, params, corpo.desugar)
  
  // Expressões atômicas
  val simple: Parser[interp.Exp] = for {
    tok <- peek
    e <- tok match {
      case NUM(n) => skip(interp.Num(n))
      case ID(s) => for {
        v <- skip(interp.Var(s))
        tok <- peek
        call <- if (tok == OP('(')) args
                else unit(None) 
      } yield call match {
        case Some(args) => interp.Ap1(s, args)
        case None => v
      }
      case TRUE() => skip(interp.True())
      case FALSE() => skip(interp.False())
      case NIL() => skip(interp.Nil())
      case IF() => parseIf
      case LET() => let
      case LETREC() => letrec
      case REC() => rec
      case FUN() => anonFun
      case REF() => for {
        _ <- token(REF())
        _ <- token(OP('('))
        e <- exp
        _ <- token(OP(')'))
      } yield interp.Ref(e)
      case DEREF() => for {
        _ <- token(DEREF())
        _ <- token(OP('('))
        e <- exp
        _ <- token(OP(')'))
      } yield interp.Deref(e)
      case OP('!') => for {
        _ <- token(OP('!'))
        e <- simple
      } yield interp.Nao(e)
      case OP('-') => for {
        _ <- token(OP('-'))
        e <- simple
      } yield interp.Menos(e)
      case OP('(') => for {
        _ <- token(OP('('))
        e <- exp
        _ <- token(OP(')'))
        tok <- peek
        call <- if (tok == OP('(')) args
                else unit(None) 
      } yield call match {
        case Some(args) => interp.Ap(e, args)
        case None => e
      }
      case _ => parseError("expressão")
    }
  } yield e
  
  // Lista de argumentos (entre parênteses) de uma chamada
  val args: Parser[Option[List[interp.Exp]]] = for {
    _ <- token(OP('('))
    tok <- peek
    list <- if (tok == OP(')')) unit(List())
            else listOf(exp)
    _ <- token(OP(')'))
  } yield Some(list)

  // Expressões
  val exp: Parser[interp.Exp] = for {
    e1 <- simple
    e <- expPrec(e1, 0)
  } yield e

  // Expressões binárias
  def expPrec(first: interp.Exp, limit: Int): Parser[interp.Exp] = for {
    left <- unit(first)
    op <- peek
    res <- prec.get(op) match {
      case Some((pleft, pright)) => if (pleft > limit) for {
                                             _ <- token(op)
                                             next <- simple
                                             right <- expPrec(next, pright)
                                             e <- expPrec(binExp(op)(left, right), limit)
                                           } yield e
                     else unit(left)
      case None => unit(left)
    }
  } yield res

  // Funções anônimas
  val anonFun: Parser[interp.Fun] = for {
    _ <- token(FUN())
    _ <- token(OP('('))
    tok <- peek
    params <- if (tok == OP(')')) unit(List[String]()) 
              else listOf(id)
    _ <- token(OP(')'))
    corpo <- exp
    _ <- token(END())
  } yield interp.Fun(params, corpo.desugar)

  // Expressão if
  val parseIf: Parser[interp.If] = for {
    _ <- token(IF())
    cond <- exp
    _ <- token(THEN())
    ethen <- exp
    _ <- token(ELSE())
    eelse <- exp
    _ <- token(END())
  } yield interp.If(cond, ethen, eelse)
  
  // Let e let múltiplo
  val let: Parser[interp.Exp] = for {
    _ <- token(LET())
    bindings <- listOf(binding)
    _ <- token(IN())
    corpo <- exp
    _ <- token(END())
  } yield bindings match {
    case (nome, e) :: Nil => interp.Let(nome, e, corpo)
    case _ => interp.LetM(bindings, corpo)
  }
  
  val letrec: Parser[interp.Exp] = for {
    _ <- token(LETREC())
    (nome, e) <- binding
    _ <- token(IN())
    corpo <- exp
    _ <- token(END())
  } yield interp.Let(nome, interp.Rec(nome, e), corpo)
    
  // Atribuição do let
  val binding: Parser[(String, interp.Exp)] = for {
    nome <- id
    _ <- token(OP('='))
    e <- exp
  } yield (nome, e)
  
  // Lista de funções do top-level
  val funs: Parser[Map[String, interp.Fun1]] = for {
    tok <- peek
    fs <- if (tok == FUN()) for {
                              f <- fun
                              fs <- funs
                            } yield fs + (f.nome -> f)
          else unit(Map[String, interp.Fun1]())
  } yield fs
  
  val rec: Parser[interp.Rec] = for {
    _ <- token(REC())
    par <- binding
  } yield par match {
    case (nome, e) => interp.Rec(nome, e)
  }
  
  // Programa
  val prog: Parser[interp.Prog] = for {
    fs <- funs
    e <- exp
    _ <- token(EOF())
  } yield interp.Prog(fs, e.desugar)
  
  def parseProg(input: Seq[Char]): interp.Prog = {
    val tokens = scanner.scan(input)
    val (p, _) = prog.parse(tokens)
    p
  }
  
  def parseExp(input: Seq[Char]): interp.Exp = {
    val tokens = scanner.scan(input)
    val (interp.Prog(_, e), _) = prog.parse(tokens)
    e
  }

  def parseProg(input: String): interp.Prog = parseProg(input.toSeq)

  def parseExp(input: String): interp.Exp = parseExp(input.toSeq)
  
  def parseFile(filename: String): interp.Prog =
    parseProg(scala.io.Source.fromFile(filename).toSeq)
}

